@@ -3,20 +3,21 @@ from typing import Optional, Union, List, Callable
 import numpy as np
 
 from . import vamp_score
-from ..base import Model, Transformer
+from ..base import Model, Transformer, Observable
 from ..basis import Identity, Concatenation
 from ..covariance import CovarianceModel, WhiteningTransform
 from ..numeric import is_diagonal_matrix
 from ..util.decorators import cached_property
+from ..util.validation import ck_test
 
 
-class KoopmanModel(Model, Transformer):
-    r""" Model which contains a finite-dimensional Koopman operator (or approximation thereof).
+class TransferOperatorModel(Model, Transformer):
+    r""" Model which contains a finite-dimensional transfer operator (or approximation thereof).
     It describes the temporal evolution of observable space, i.e.,
 
     .. math:: \mathbb{E}[g(x_{t+\tau})] = K^\top \mathbb{E}[f(x_t)],
 
-    where :math:`K\in\mathbb{R}^{n\times m}` is the Koopman operator, :math:`x_t` the system's state at time :math:`t`,
+    where :math:`K\in\mathbb{R}^{n\times m}` is the transfer operator, :math:`x_t` the system's state at time :math:`t`,
     and :math:`f` and :math:`g` observables of the system's state.
 
     Parameters
@@ -39,7 +40,7 @@ class KoopmanModel(Model, Transformer):
 
     @property
     def operator(self) -> np.ndarray:
-        r""" The operator :math:`K` so that :math:`\mathbb{E}[g(x_{t+\tau}] = K^\top \mathbb{E}[f(x_t)]` in transformed
+        r""" The operator :math:`K` so that :math:`\mathbb{E}[g(x_{t+\tau})] = K^\top \mathbb{E}[f(x_t)]` in transformed
         bases.
         """
         return self._koopman_matrix
@@ -124,10 +125,10 @@ class KoopmanModel(Model, Transformer):
         transformed_data : (T,m) ndarray
             The transformed data.
         """
-        raise NotImplementedError()
+        return self.instantaneous_obs(data)
 
 
-class CovarianceKoopmanModel(KoopmanModel):
+class CovarianceKoopmanModel(TransferOperatorModel):
     r"""A type of Koopman model :math:`\mathbb{E}[g(x_{t+\tau})] = K^\top \mathbb{E}[f(x_{t})]`
     which was obtained through diagonalization of covariance matrices. This leads to
     a Koopman operator which is a diagonal matrix and can be used to project onto specific processes of the system.
@@ -144,7 +145,7 @@ class CovarianceKoopmanModel(KoopmanModel):
     :class:`TICA <deeptime.decomposition.TICA>`.
 
     For a description of parameters `operator`, `basis_transform_forward`, `basis_transform_backward`,
-    and `output_dimension`: please see :class:`KoopmanModel`.
+    and `output_dimension`: please see :class:`TransferOperatorModel`.
 
     Parameters
     ----------
@@ -168,9 +169,9 @@ class CovarianceKoopmanModel(KoopmanModel):
         with norms <= epsilon were cut off. The remaining number of eigenvalues together with the value of `dim`
         define the effective output dimension.
     instantaneous_obs : Callable, optional, default=identity
-        Transforms the current state :math:`x_t` to :math:`\chi_0(x_t)`. Defaults to `\chi_0(x) = x`.
+        Transforms the current state :math:`x_t` to :math:`\chi_0(x_t)`. Defaults to :math:`\chi_0(x) = x`.
     timelagged_obs : Callable, optional, default=identity
-        Transforms the future state :math:`x_{t+\tau}` to :math:`\chi_1(x_{t+\tau})`. Defaults to `\chi_1(x) = x`.
+        Transforms the future state :math:`x_{t+\tau}` to :math:`\chi_1(x_{t+\tau})`. Defaults to :math:`\chi_1(x) = x`.
     """
 
     def __init__(self, instantaneous_coefficients, singular_values, timelagged_coefficients, cov,
@@ -350,9 +351,8 @@ class CovarianceKoopmanModel(KoopmanModel):
 
     @var_cutoff.setter
     def var_cutoff(self, value):
-        if value <= 0. or float(value) > 1.0:
-            raise ValueError("VAMP: Invalid dimension parameter, if it is given in terms of a floating point, "
-                             "can only be in the interval (0, 1].")
+        assert 0 < value <= 1., "Invalid dimension parameter, if it is given in terms of a variance cutoff, " \
+                                "it can only be in the interval (0, 1]."
         self._var_cutoff = value
         self._update_output_dimension()
 
@@ -371,11 +371,10 @@ class CovarianceKoopmanModel(KoopmanModel):
 
     @dim.setter
     def dim(self, value: Optional[int]):
-        if isinstance(value, int):
-            if value <= 0:
-                # first test against Integral as `isinstance(1, Real)` also evaluates to True
-                raise ValueError("VAMP: Invalid dimension parameter, if it is given in terms of the "
-                                 "dimension (integer), must be positive.")
+        if isinstance(value, int) and value <= 0:
+            # first test against Integral as `isinstance(1, Real)` also evaluates to True
+            raise ValueError("VAMP: Invalid dimension parameter, if it is given in terms of the "
+                             "dimension (integer), must be positive.")
         self._dim = value
         self._update_output_dimension()
 
@@ -526,7 +525,7 @@ class CovarianceKoopmanModel(KoopmanModel):
 
         Notes
         -----
-        A "future expectation" of a observable g is the average of g computed
+        A "future expectation" of an observable :math:`g` is the average of :math:`g` computed
         over a time window that has the same total length as the input data
         from which the Koopman operator was estimated but is shifted
         by ``lag_multiple*tau`` time steps into the future (where tau is the lag
@@ -554,7 +553,7 @@ class CovarianceKoopmanModel(KoopmanModel):
 
 
         A model prediction of time-lagged covariances between the
-        observable f and the statistic g at a lag-time of ``lag_multiple*tau``
+        observable :math:`f` and the statistic :math:`g` at a lag-time of ``lag_multiple*tau``
         is computed with the equation:
 
         .. math::
@@ -575,23 +574,29 @@ class CovarianceKoopmanModel(KoopmanModel):
         m_0 = self.mean_0
         m_t = self.mean_t
 
+        integer_multiple = np.issubdtype(type(lag_multiple), np.integer)
+        dtype = float if integer_multiple else complex
         if lag_multiple == 1:
             P = S
         else:
-            p = np.zeros((dim + 1, dim + 1))
+            p = np.zeros((dim + 1, dim + 1), dtype=dtype)
             p[0, 0] = 1.0
             p[1:, 0] = U.T.dot(m_t - m_0)
             p[1:, 1:] = U.T.dot(self.cov_tt).dot(V)
-            P = np.linalg.matrix_power(S.dot(p), lag_multiple - 1).dot(S)
+            if integer_multiple:
+                P = np.linalg.matrix_power(S.dot(p), lag_multiple - 1).dot(S)
+            else:
+                from scipy.linalg import fractional_matrix_power
+                P = fractional_matrix_power(S.dot(p), lag_multiple - 1).dot(S)
 
-        Q = np.zeros((observables.shape[1], dim + 1))
+        Q = np.zeros((observables.shape[1], dim + 1), dtype=dtype)
         if not observables_mean_free:
             Q[:, 0] = observables.T.dot(m_t)
         Q[:, 1:] = observables.T.dot(self.cov_tt).dot(V)
 
         if statistics is not None:
             # compute covariance
-            R = np.zeros((statistics.shape[1], dim + 1))
+            R = np.zeros((statistics.shape[1], dim + 1), dtype=dtype)
             if not statistics_mean_free:
                 R[:, 0] = statistics.T.dot(m_0)
             R[:, 1:] = statistics.T.dot(self.cov_00).dot(U)
@@ -641,7 +646,13 @@ class CovarianceKoopmanModel(KoopmanModel):
             raise ValueError("This is only meaningful for real singular values.")
         if lagtime is None:
             lagtime = self._cov.lagtime
-        return - lagtime / np.log(np.abs(self.singular_values[:k]))
+        return -lagtime / np.log(np.abs(self.singular_values[:k]))
+
+    @property
+    def lagtime(self):
+        r""" The lagtime corresponding to this model. See also
+        :meth:`CovarianceModel.lagtime <deeptime.covariance.CovarianceModel.lagtime>`. """
+        return self._cov.lagtime
 
     @property
     def feature_component_correlation(self):
@@ -663,4 +674,107 @@ class CovarianceKoopmanModel(KoopmanModel):
             feature and a column for each component.
         """
         feature_sigma = np.sqrt(np.diag(self.cov_00))
-        return np.dot(self.cov_00, self.singular_vectors_left[:, : self.output_dimension]) / feature_sigma[:, None]
\ No newline at end of file
+        return np.dot(self.cov_00, self.singular_vectors_left[:, : self.output_dimension]) / feature_sigma[:, None]
+
+    def ck_test(self, models, test_model=None, include_lag0=True, n_observables=None,
+                observables='phi', statistics='psi', progress=None):
+        r"""Returns a Chapman-Kolmogorov validator based on this estimator and a test model.
+
+        Parameters
+        ----------
+        models : list of models
+            Multiple models with different lagtimes to test against.
+        test_model : CovarianceKoopmanModel, optional, default=None
+            The model that is tested. If not provided, uses this estimator's encapsulated model.
+        include_lag0 : bool, optional, default=True
+            Whether to include lagtime 0.
+        n_observables : int, optional, default=None
+            Limit the number of default observables (and of default statistics) to this number.
+            Only used if `observables` are None or `statistics` are None.
+        observables : (input_dimension, n_observables) ndarray
+            Coefficients that express one or multiple observables in
+            the basis of the input features.
+        statistics : (input_dimension, n_statistics) ndarray
+            Coefficients that express one or multiple statistics in the basis of the input features.
+        progress : ProgressBar, optional, default=None
+            Optional progress bar, tested for tqdm.
+
+        Returns
+        -------
+        test : deeptime.util.validation.ChapmanKolmogorovTest
+            The test results
+
+        See Also
+        --------
+        :meth:`ck_test <deeptime.util.validation.ck_test>`
+
+        Notes
+        -----
+        This method computes two sets of time-lagged covariance matrices
+
+        * estimates at higher lag times :
+
+          .. math::
+
+              \left\langle \mathbf{K}(n\tau)g_{i},f_{j}\right\rangle_{\rho_{0}}
+
+          where :math:`\rho_{0}` is the empirical distribution implicitly defined
+          by all data points from time steps 0 to T-tau in all trajectories,
+          :math:`\mathbf{K}(n\tau)` is a rank-reduced Koopman matrix estimated
+          at the lag-time n*tau and g and f are some functions of the data.
+
+        * predictions at higher lag times :
+
+          .. math::
+
+              \left\langle \mathbf{K}^{n}(\tau)g_{i},f_{j}\right\rangle_{\rho_{0}}
+
+          where :math:`\mathbf{K}^{n}` is the n'th power of the rank-reduced
+          Koopman matrix contained in self.
+
+        The Champan-Kolmogorov test is to compare the predictions to the estimates.
+        """
+        test_model = self if test_model is None else test_model
+        if n_observables is not None:
+            if n_observables > test_model.dim:
+                import warnings
+                warnings.warn('Selected singular functions as observables but dimension '
+                              'is lower than requested number of observables.')
+                n_observables = test_model.dim
+        else:
+            n_observables = test_model.dim
+
+        if isinstance(observables, str) and observables == 'phi':
+            observables = test_model.singular_vectors_right[:, :n_observables]
+            observables_mean_free = True
+        else:
+            observables_mean_free = False
+
+        if isinstance(statistics, str) and statistics == 'psi':
+            statistics = test_model.singular_vectors_left[:, :n_observables]
+            statistics_mean_free = True
+        else:
+            statistics_mean_free = False
+        observable = KoopmanObservable(observables, statistics, observables_mean_free, statistics_mean_free)
+        return ck_test(models, observable, test_model=test_model, include_lag0=include_lag0, err_est=False,
+                       progress=progress)
+
+
+class KoopmanObservable(Observable):
+    def __init__(self, observables, statistics, observables_mean_free, statistics_mean_free):
+        self.observables = observables
+        self.statistics = statistics
+        self.observables_mean_free = observables_mean_free
+        self.statistics_mean_free = statistics_mean_free
+
+    def __call__(self, model, mlag=1, **kw):
+        if mlag == 0 or model is None:
+            # todo for lag time 0 we return a matrix of nan, until the correct solution is implemented
+            if self.statistics is None:
+                return np.full((self.observables.shape[1],), fill_value=np.nan)
+            else:
+                return np.full((self.observables.shape[1], self.statistics.shape[1]), fill_value=np.nan)
+        else:
+            return model.expectation(statistics=self.statistics, observables=self.observables, lag_multiple=mlag,
+                                     statistics_mean_free=self.statistics_mean_free,
+                                     observables_mean_free=self.observables_mean_free)
