@@ -1,13 +1,12 @@
-from typing import Optional, Union, Callable, Tuple, List
+from typing import Optional, Union, Callable, Tuple
 
 import numpy as np
 import torch
 import torch.nn as nn
-from torch.utils.data import Dataset
 
-from ...base import Model, Transformer
-from ...base_torch import DLEstimator
-from ...util.torch import map_data
+from ...base import Model, Transformer, EstimatorTransformer
+from ...base_torch import DLEstimatorMixin
+from ...util.torch import map_data, eigh, multi_dot, disable_TF32
 
 
 def symeig_reg(mat, epsilon: float = 1e-6, mode='regularize', eigenvectors=True) \
@@ -39,7 +38,7 @@ def symeig_reg(mat, epsilon: float = 1e-6, mode='regularize', eigenvectors=True)
         mat = mat + epsilon * identity
 
     # Calculate eigvalues and potentially eigvectors
-    eigval, eigvec = torch.symeig(mat, eigenvectors=True)
+    eigval, eigvec = eigh(mat)
 
     if eigenvectors:
         eigvec = eigvec.transpose(0, 1)
@@ -56,13 +55,7 @@ def symeig_reg(mat, epsilon: float = 1e-6, mode='regularize', eigenvectors=True)
     elif mode == 'clamp':
         eigval = torch.clamp_min(eigval, min=epsilon)
 
-    else:
-        raise RuntimeError("Invalid mode! Should have been caught by the assertion.")
-
-    if eigenvectors:
-        return eigval, eigvec
-    else:
-        return eigval, eigvec
+    return eigval, eigvec
 
 
 def sym_inverse(mat, epsilon: float = 1e-6, return_sqrt=False, mode='regularize'):
@@ -96,7 +89,7 @@ def sym_inverse(mat, epsilon: float = 1e-6, return_sqrt=False, mode='regularize'
     else:
         diag = torch.diag(1. / eigval)
 
-    return torch.chain_matmul(eigvec.t(), diag, eigvec)
+    return multi_dot([eigvec.t(), diag, eigvec])
 
 
 sym_inverse.valid_modes = ('trunc', 'regularize', 'clamp')
@@ -134,7 +127,7 @@ def koopman_matrix(x: torch.Tensor, y: torch.Tensor, epsilon: float = 1e-6, mode
         c00, c0t, ctt = covariances(x, y, remove_mean=True)
     c00_sqrt_inv = sym_inverse(c00, return_sqrt=True, epsilon=epsilon, mode=mode)
     ctt_sqrt_inv = sym_inverse(ctt, return_sqrt=True, epsilon=epsilon, mode=mode)
-    return torch.chain_matmul(c00_sqrt_inv, c0t, ctt_sqrt_inv).t()
+    return multi_dot([c00_sqrt_inv, c0t, ctt_sqrt_inv]).t()
 
 
 def covariances(x: torch.Tensor, y: torch.Tensor, remove_mean: bool = True):
@@ -184,7 +177,6 @@ def covariances(x: torch.Tensor, y: torch.Tensor, remove_mean: bool = True):
 
 valid_score_methods = ('VAMP1', 'VAMP2', 'VAMPE')
 
-
 def vamp_score(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', epsilon: float = 1e-6, mode='trunc'):
     r"""Computes the VAMP score based on data and corresponding time-shifted data.
 
@@ -207,10 +199,10 @@ def vamp_score(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', ep
         The score. It contains a contribution of :math:`+1` for the constant singular function since the
         internally estimated Koopman operator is defined on a decorrelated basis set.
     """
-    if method not in valid_score_methods:
-        raise ValueError(f"Invalid method '{method}', supported are {valid_score_methods}")
-    assert data.shape == data_lagged.shape
-
+    assert method in valid_score_methods, f"Invalid method '{method}', supported are {valid_score_methods}"
+    assert data.shape == data_lagged.shape, f"Data and data_lagged must be of same shape but were {data.shape} " \
+                                            f"and {data_lagged.shape}."
+    out = None
     if method == 'VAMP1':
         koopman = koopman_matrix(data, data_lagged, epsilon=epsilon, mode=mode)
         out = torch.norm(koopman, p='nuc')
@@ -221,7 +213,7 @@ def vamp_score(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', ep
         c00, c0t, ctt = covariances(data, data_lagged, remove_mean=True)
         c00_sqrt_inv = sym_inverse(c00, epsilon=epsilon, return_sqrt=True, mode=mode)
         ctt_sqrt_inv = sym_inverse(ctt, epsilon=epsilon, return_sqrt=True, mode=mode)
-        koopman = torch.chain_matmul(c00_sqrt_inv, c0t, ctt_sqrt_inv).t()
+        koopman = multi_dot([c00_sqrt_inv, c0t, ctt_sqrt_inv]).t()
 
         u, s, v = torch.svd(koopman)
         mask = s > epsilon
@@ -235,14 +227,11 @@ def vamp_score(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', ep
         s = torch.diag(s)
 
         out = torch.trace(
-            2. * torch.chain_matmul(s, u_t, c0t, v)
-            - torch.chain_matmul(s, u_t, c00, u, s, v_t, ctt, v)
+            2. * multi_dot([s, u_t, c0t, v]) - multi_dot([s, u_t, c00, u, s, v_t, ctt, v])
         )
-    else:
-        raise RuntimeError("This should have been caught earlier.")
+    assert out is not None
     return 1 + out
 
-
 def vampnet_loss(data: torch.Tensor, data_lagged: torch.Tensor, method='VAMP2', epsilon: float = 1e-6,
                  mode: str = 'trunc'):
     r"""Loss function that can be used to train VAMPNets. It evaluates as :math:`-\mathrm{score}`. The score
@@ -283,22 +272,9 @@ class VAMPNetModel(Transformer, Model):
         elif dtype == np.float64:
             self._lobe = self._lobe.double()
             self._lobe_timelagged = self._lobe_timelagged.double()
-        self._instantaneous = True
         self._dtype = dtype
         self._device = device
 
-    @property
-    def instantaneous(self) -> bool:
-        r""" Whether to use the instantaneous lobe or the time-shifted one.
-
-        :type: bool
-        """
-        return self._instantaneous
-
-    @instantaneous.setter
-    def instantaneous(self, value: bool):
-        self._instantaneous = value
-
     @property
     def lobe(self) -> nn.Module:
         r""" The instantaneous lobe.
@@ -319,8 +295,24 @@ class VAMPNetModel(Transformer, Model):
         """
         return self._lobe_timelagged
 
-    def transform(self, data, **kwargs):
-        if self.instantaneous:
+    def transform(self, data, instantaneous: bool = True, **kwargs):
+        r""" Transforms data through the instantaneous or time-shifted network lobe.
+
+        Parameters
+        ----------
+        data : numpy array or torch tensor
+            The data to transform.
+        instantaneous : bool, default=True
+            Whether to use the instantaneous lobe or the time-shifted lobe for transformation.
+        **kwargs
+            Ignored kwargs for api compatibility.
+
+        Returns
+        -------
+        transform : array_like
+            List of numpy array or numpy array containing transformed data.
+        """
+        if instantaneous:
             self._lobe.eval()
             net = self._lobe
         else:
@@ -332,7 +324,7 @@ class VAMPNetModel(Transformer, Model):
         return out if len(out) > 1 else out[0]
 
 
-class VAMPNet(DLEstimator, Transformer):
+class VAMPNet(EstimatorTransformer, DLEstimatorMixin):
     r""" Implementation of VAMPNets. :footcite:`mardt2018vampnets`
     These networks try to find an optimal featurization of data based on a VAMP score :footcite:`wu2020variational`
     by using neural networks as featurizing transforms which are equipped with a loss that is the negative VAMP score.
@@ -365,8 +357,6 @@ class VAMPNet(DLEstimator, Transformer):
         see :meth:`sym_inverse`.
     dtype : dtype, default=np.float32
         The data type of the modules and incoming data.
-    shuffle : bool, default=True
-        Whether to shuffle data during training after each epoch.
 
     See Also
     --------
@@ -381,14 +371,13 @@ class VAMPNet(DLEstimator, Transformer):
     def __init__(self, lobe: nn.Module, lobe_timelagged: Optional[nn.Module] = None,
                  device=None, optimizer: Union[str, Callable] = 'Adam', learning_rate: float = 5e-4,
                  score_method: str = 'VAMP2', score_mode: str = 'regularize', epsilon: float = 1e-6,
-                 dtype=np.float32, shuffle: bool = True):
+                 dtype=np.float32):
         super().__init__()
         self.lobe = lobe
         self.lobe_timelagged = lobe_timelagged
         self.score_method = score_method
         self.score_mode = score_mode
         self._step = 0
-        self.shuffle = shuffle
         self._epsilon = epsilon
         self.device = device
         self.learning_rate = learning_rate
@@ -441,9 +430,8 @@ class VAMPNet(DLEstimator, Transformer):
 
     @score_method.setter
     def score_method(self, value: str):
-        if value not in valid_score_methods:
-            raise ValueError(f"Tried setting an unsupported scoring method '{value}', "
-                             f"available are {valid_score_methods}.")
+        assert value in valid_score_methods, f"Tried setting an unsupported scoring method '{value}', " \
+                                             f"available are {valid_score_methods}."
         self._score_method = value
 
     @property
@@ -553,14 +541,15 @@ class VAMPNet(DLEstimator, Transformer):
         score : torch.Tensor
             The value of the score.
         """
-        self.lobe.eval()
-        self.lobe_timelagged.eval()
+        with disable_TF32():
+            self.lobe.eval()
+            self.lobe_timelagged.eval()
 
-        with torch.no_grad():
-            val = self.lobe(validation_data[0])
-            val_t = self.lobe_timelagged(validation_data[1])
-            score_value = vamp_score(val, val_t, method=self.score_method, mode=self.score_mode, epsilon=self.epsilon)
-            return score_value
+            with torch.no_grad():
+                val = self.lobe(validation_data[0])
+                val_t = self.lobe_timelagged(validation_data[1])
+                score_value = vamp_score(val, val_t, method=self.score_method, mode=self.score_mode, epsilon=self.epsilon)
+                return score_value
 
     def fit(self, data_loader: torch.utils.data.DataLoader, n_epochs=1, validation_loader=None,
             train_score_callback: Callable[[int, torch.Tensor], None] = None,
@@ -599,44 +588,24 @@ class VAMPNet(DLEstimator, Transformer):
         self._step = 0
 
         # and train
-        for epoch in progress(range(n_epochs), desc="VAMPNet epoch", total=n_epochs, leave=False):
-            for batch_0, batch_t in data_loader:
-                self.partial_fit((batch_0.to(device=self.device), batch_t.to(device=self.device)),
-                                 train_score_callback=train_score_callback)
-            if validation_loader is not None:
-                with torch.no_grad():
-                    scores = []
-                    for val_batch in validation_loader:
-                        scores.append(
-                            self.validate((val_batch[0].to(device=self.device), val_batch[1].to(device=self.device)))
-                        )
-                    mean_score = torch.mean(torch.stack(scores))
-                    self._validation_scores.append((self._step, mean_score.item()))
-                    if validation_score_callback is not None:
-                        validation_score_callback(self._step, mean_score)
+        with disable_TF32():
+            for _ in progress(range(n_epochs), desc="VAMPNet epoch", total=n_epochs, leave=False):
+                for batch_0, batch_t in data_loader:
+                    self.partial_fit((batch_0.to(device=self.device), batch_t.to(device=self.device)),
+                                     train_score_callback=train_score_callback)
+                if validation_loader is not None:
+                    with torch.no_grad():
+                        scores = []
+                        for val_batch in validation_loader:
+                            scores.append(
+                                self.validate((val_batch[0].to(device=self.device), val_batch[1].to(device=self.device)))
+                            )
+                        mean_score = torch.mean(torch.stack(scores))
+                        self._validation_scores.append((self._step, mean_score.item()))
+                        if validation_score_callback is not None:
+                            validation_score_callback(self._step, mean_score)
         return self
 
-    def transform(self, data, instantaneous: bool = True, **kwargs):
-        r""" Transforms data through the instantaneous or time-shifted network lobe.
-
-        Parameters
-        ----------
-        data : numpy array or torch tensor
-            The data to transform.
-        instantaneous : bool, default=True
-            Whether to use the instantaneous lobe or the time-shifted lobe for transformation.
-        **kwargs
-            Ignored kwargs for api compatibility.
-
-        Returns
-        -------
-        transform : array_like
-            List of numpy array or numpy array containing transformed data.
-        """
-        model = self.fetch_model()
-        model.instantaneous = instantaneous
-        return model.transform(data, **kwargs)
-
     def fetch_model(self) -> VAMPNetModel:
         r""" Yields the current model. """
         from copy import deepcopy
